UNIT HIDinterface;
{
  USB: Human Interface Device class interface file:
  Basic calls to communicate with an HID device

  When using this unit/library make sure the next "standard"
  Windows DLL's are installed on the target system:
     - HID.DLL
     - SetupAPI.DLL

  Created:     16-04-2012 by J.W. Beunders
  Last update: 17-08-2012 by J.W. Beunders (joop@beunders.be)

  This code may be used freely. The author would appreciate it
  when he is mentioned as original creator of the interface.
  Also he would be very grateful if he is notified when changes
  and / or improvements are made,
}

INTERFACE

CONST
{
  he number below indicates the maximum number of HID devices the scan
  function can return.
}
  HIDmaxDevices = 31;

TYPE
{
  Basic HID data buffer. This the maximum size of an HID buffer.
  The actual size for a specific device should be cheched
  before any communication is started
}
  THIDbuffer        = Array[0..63] of Byte;

{
  The interface includes a thread which tries to read data from
  the active HID device. When a callback procdure is assigned in
  the main program, all available data is automatically returned
  to the application in this procedure
}
  THIDcallBack      = Procedure( Data : THIDbuffer);  stdcall;

{
  When a USB device is attached or removed from the system, the
  system will generate a message. By specifying a callback procedure
  in the main program, the event can be used to take some action,
  for instance to check if the active device is still present.
}
  TUSBcallBack      = Procedure; stdcall;

{
  An HID device is specified by the structure shown below. Apart from
  the VID/PID combination, the beuffersize is of great importance
  as that specifies how much data can be transferred in a single vcall.
}
  THIDdeviceInfo = Record
    SymLink            : ShortString;  // symbolic link from registry
    BufferSize         : Word;  // size of the data buffer for sending/receiving data
    Handle             : THandle;  // handle which is returned when device is opened
    VID                : Word;  // Vendor ID code
    PID                : Word;  // Product ID code
    VersionNumber      : Word;  // Vrsion information
    ManufacturerString : ShortString;  // Textual manufacurer information
    ProductString      : ShortString;  // Textual product information
    SerialNumberString : ShortString;  // Textual serial number information
  end;
              
{
  The structure below contains information of all HID devices after a call
  to ScanForHIDdevices.
}
  THIDDeviceList = Array[0..HIDmaxDevices] of THIDdeviceInfo;

{
    assign procedure to call when a usb device is inserted or removed
}
Procedure USBsetEventHandler( Callback : TUSBcallBack);  stdcall;

{
  assign a user procedure which will receive data when HID device has send it
}
Procedure HIDsetEventHandler( CallBack : THIDcallback);  stdcall;

{
  Scan all available HID Devices.
  When Target VID/PID equals 0 all HID entries from device manager are
  returned. To check the presence of a particular device, enter the
  required VID/PID information
}
Procedure HIDscanForDevices( Var DeviceList : THIDDeviceList;
                             Var NumDevices : Byte;
                                 TargetVID,
                                 TargetPID  : Word);  stdcall;

{
  Get a Handle to the device (open it)
}
Function HIDopenDevice( Var DeviceInfo : THIDdeviceInfo) : Boolean;  stdcall;

{
  Close Device
}
Function HIDcloseDevice( Var DeviceInfo : THIDdeviceInfo) : Boolean;  stdcall;

{
  Write NrOfBytes to HID device
}
Function HIDwriteDevice( Var DeviceInfo   : THIDdeviceInfo;
                             Data         : THIDbuffer) : Boolean;  stdcall;

{
  Read data from device with a time-out.
  When succesfull NrOfBytes holds the actual number of bytes
}
Function HIDreadDevice( Var DeviceInfo : THIDdeviceInfo;
                        Var Data       : THIDbuffer ) : Boolean;  stdcall;


IMPLEMENTATION
USES
  Windows,
  Classes,
  Messages,
  HID,
  SetupApi,
  ModuleLoader;

TYPE

  THIDreport = Record
    ReportID : Byte;  // always 0
    Data     : THIDbuffer;  // communication buffer
  end;

  THIDthread = Class(TThread)
    Private
    Public
      Constructor Create;
      Procedure Execute; override;
      Procedure Free;
  end;

  TUSBhandler = Class
    Private
      fMsgHandlerHWND : HWND;
      procedure USBchanger(var Msg: TMessage);
    Public
      Constructor Create(AOwner: TComponent);
      Destructor Destroy; override;
  end;

CONST
  HID_TIMEOUT                    = 200;  // Timeout value in ms for low level HID transfer
  HIDUSB_COUNTOFINTERRUPTBUFFERS = 64;   // Count of buffers for interrupt data

VAR
  ActiveDevice                   : THIDdeviceInfo;
  HIDthread                      : THIDthread        = Nil;
  HIDcallBack                    : THIDcallBack      = Nil;
  USBhandler                     : TUSBhandler       = Nil;
  USBeventHandler                : TUSBcallBack      = Nil;

Procedure ClearDeviceInfo( Var Device : THIDdeviceInfo);
Begin
  With Device do
    begin
      SymLink            := '';   
      BufferSize         := 0;    
      Handle             := INVALID_HANDLE_VALUE;
      VID                := 0;
      PID                := 0;    
      VersionNumber      := 0;
      ManufacturerString := '';
      ProductString      := '';   
      SerialNumberString := '';
    end;
End;


Procedure HIDscanForDevices( Var DeviceList : THIDDeviceList;
                             Var NumDevices : Byte;
                                 TargetVID,
                                 TargetPID  : Word);
Var
  HID_GUIid : TGUID;
  spdid     : TSPDeviceInterfaceData;
  pSpDidd   : PSPDEVICEINTERFACEDETAILDATAA;
  spddd     : TSPDevInfoData;
  HIDinfo   : HDEVINFO;
  CurIdx    : Integer;
  dwSize    : DWord;
  DevHandle : THandle;
  HidAttrs  : THIDDAttributes;
  Info      : THIDdeviceInfo;

  Function GetHidDeviceInfo( Symlink : PAnsiChar) : THIDdeviceInfo;
  Var
    pStr          : pWideChar;
    PreparsedData : PHIDPPreparsedData;
    HidCaps       : THIDPCaps;
  Begin
    FillChar(Result, SizeOf(Result), 0);
    Result.SymLink := SymLink + #0;
    GetMem(pstr, 512);
    DevHandle := CreateFileW( PWIDECHAR( WideString( Symlink)),
                              GENERIC_READ or GENERIC_WRITE,
                              FILE_SHARE_READ OR FILE_SHARE_WRITE,
                              nil,
                              OPEN_EXISTING,
                              0,
                              0);
    If (DevHandle <> INVALID_HANDLE_VALUE) then
      begin
        If HidD_GetAttributes( DevHandle,
                               HidAttrs) then
          begin
            Result.VID           := HidAttrs.VendorID;
            Result.PID           := HidAttrs.ProductID;
            Result.VersionNumber := HidAttrs.VersionNumber;
          end;
        If HidD_GetManufacturerString( DevHandle,
                                       pStr,
                                       512) then Result.ManufacturerString := String(pStr);
        If HidD_GetProductString( DevHandle,
                                  pStr,
                                  512) then Result.ProductString := String(pStr);
        If HidD_GetSerialNumberString( DevHandle,
                                       pStr,
                                       512) then Result.SerialNumberString := String(pStr)
                                       else Result.SerialNumberString := '---';
                                       
        { Set Input buffer size }
        HidD_SetNumInputBuffers( DevHandle,
                                 HIDUSB_COUNTOFINTERRUPTBUFFERS);
        { Get capabilities }
        HidD_GetPreparsedData( DevHandle,
                               PreparsedData);

        If ( PreparsedData <> 0) then
          begin
            HidP_GetCaps( PreparsedData,
                          HidCaps);
            Result.BufferSize := HidCaps.OutputReportByteLength;
          end
        else Result.BufferSize := 0;
        CloseHandle( DevHandle);
      end;
    FreeMem(pStr);
  End;

Begin  // HIDscanForDevices
  // clear entire device list
  For CurIdx := 0 to HIDmaxDevices do ClearDeviceInfo( DeviceList[ CurIdx]);
  // assume no devices are present
  NumDevices := 0;
  // Get GUID of hid class
  HidD_GetHidGuid(HID_GUIid);
  // Get a list of devices belonging to HID class
  HIDinfo := SetupDiGetClassDevs( @HID_GUIid,
                                  nil,
                                  GetDesktopWindow(),
                                  DIGCF_DEVICEINTERFACE or DIGCF_PRESENT);
  // Go through list of devices
  If (THandle( HIDinfo) <> INVALID_HANDLE_VALUE) then
    begin
      CurIdx := 0;
      spdid.cbSize := SizeOf( spdid);
      While SetupDiEnumDeviceInterfaces( HidInfo,
                                         nil,
                                         HID_GUIid,
                                         curIdx,
                                         spdid) do
        begin
          dwSize := 0;
          // Get device path for Createfile calls
          SetupDiGetDeviceInterfaceDetail( HIDinfo,
                                            @spdid,
                                            nil,
                                            dwSize,
                                            @dwSize,
                                            nil);
          If (dwSize > 0) then
            begin
              GetMem( pSpDidd,
                      dwSize);
              pSpDidd^.cbSize := SizeOf( TSPDEVICEINTERFACEDETAILDATAA);
              spddd.cbSize    := SizeOf( spddd);
              If SetupDiGetDeviceInterfaceDetail( HIDinfo,
                                                  @spdid,
                                                  pSpDidd,
                                                  dwSize,
                                                  nil,
                                                  @spddd) then
                begin
                  // Get information about the device (Vendor and Product IDs, Strings, ...)
                  FillChar( Info,
                            SizeOf( Info),
                            0);
                  Info        := GetHidDeviceInfo( @pSpDidd^.DevicePath);

                  Info.Handle := INVALID_HANDLE_VALUE;
                  // check if VID/PID match targets
                  If (Info.VID = TargetVID) AND
                     (Info.PID = TargetPID) then
                    begin
                      // Add Devices to result list
                      DeviceList[NumDevices] := Info;
                      If (NumDevices < HIDmaxDevices) then INC( NumDevices);
                    end
                  else  // list all HID devices if no target is specified

                  If (TargetVID = 0) AND
                     (TargetPID = 0) then
                    begin
                      // Add Devices to result list
                      DeviceList[NumDevices] := Info;
                      If (NumDevices < HIDmaxDevices) then INC( NumDevices);
                    end;

                end;
              FreeMem( pSpDidd);
            end;
          INC( CurIdx);
        end;
      SetupDiDestroyDeviceInfoList( HidInfo);
    end;
End;

Function HIDopenDevice( Var DeviceInfo : THIDdeviceInfo) : Boolean;
Begin
  If (DeviceInfo.Handle = INVALID_HANDLE_VALUE) then
  DeviceInfo.Handle := CreateFileW( PWIDECHAR(WideString( DeviceInfo.SymLink)),
                                    GENERIC_READ or GENERIC_WRITE OR SYNCHRONIZE,
                                    FILE_SHARE_READ OR FILE_SHARE_WRITE,
                                    nil,
                                    OPEN_EXISTING,
                                    FILE_FLAG_OVERLAPPED,
                                    0);
  Result := (DeviceInfo.Handle <> INVALID_HANDLE_VALUE);
  // create local copy of opened device to be used in read thread
  If Result then ActiveDevice := DeviceInfo
  else ClearDeviceInfo( ActiveDevice);
End;

Function HIDcloseDevice( Var DeviceInfo : THIDdeviceInfo) : Boolean;
Begin
  If (DeviceInfo.Handle <> INVALID_HANDLE_VALUE) then
    begin
      CloseHandle( DeviceInfo.Handle);
      DeviceInfo.Handle := INVALID_HANDLE_VALUE;
    end;
  Result := (DeviceInfo.Handle = INVALID_HANDLE_VALUE);
  // remove local copy of activated device
  If Result then ClearDeviceInfo( ActiveDevice);
end;

Function HIDwriteDevice( Var DeviceInfo   : THIDdeviceInfo;
                             Data         : THIDbuffer) : Boolean;
Var
  ReportEvent   : THandle;
  HidOverlapped : TOverlapped;
  bResult       : DWord;
  BytesWritten  : DWord;
  SendBuffer    : THIDreport;
Begin
  Result := FALSE;  // assume failure
  // check if device is open; if not then try to open device
  If (DeviceInfo.Handle = INVALID_HANDLE_VALUE) then HIDopenDevice( DeviceInfo);
  // check again if device is opened; abort if not
  If (DeviceInfo.Handle <> INVALID_HANDLE_VALUE) then
    begin
      // Create a new event for report capture
      ReportEvent              := CreateEvent( nil,
                                               TRUE,
                                               FALSE,
                                               nil);
      HidOverlapped.hEvent     := ReportEvent;
      HidOverlapped.Offset     := 0;
      HidOverlapped.OffsetHigh := 0;

      SendBuffer.ReportID := 0;  // load HID report ID
          // copy user data to send buffer
      SendBuffer.Data := Data;
      // send data to device
      WriteFile( DeviceInfo.Handle,
                 SendBuffer,
                 DeviceInfo.BufferSize,
                 BytesWritten,
                 @HidOverlapped);
      bResult := WaitForSingleObject( ReportEvent, HID_TIMEOUT);
      If (bResult = WAIT_TIMEOUT) OR (bResult = WAIT_ABANDONED) then
        begin
          CancelIo( DeviceInfo.Handle);
          Result := FALSE;
        end
      else Result := TRUE;
    end;
End;

Function HIDreadDevice( Var DeviceInfo : THIDdeviceInfo;
                        Var Data       : THIDbuffer) : Boolean;
Var
  ReportEvent   : THandle;
  HidOverlapped : TOverlapped;
  bResult, I    : DWord;
  BytesRead     : DWord;
  ReadBuffer    : THIDreport;
Begin
  Result := FALSE;  // assume failure
  // check if device is open; if not then try to open device
  If (DeviceInfo.Handle = INVALID_HANDLE_VALUE) then HIDopenDevice( DeviceInfo);
  // check again if device is opened; abort if not
  If (DeviceInfo.Handle <> INVALID_HANDLE_VALUE) then
    begin
      // Create a new event for report capture
      ReportEvent              := CreateEvent( nil,
                                               TRUE,
                                               FALSE,
                                               nil);
      HidOverlapped.hEvent     := ReportEvent;
      HidOverlapped.Offset     := 0;
      HidOverlapped.OffsetHigh := 0;
     // clear read buffer
      For I := 0 to DeviceInfo.BufferSize - 2 do ReadBuffer.Data[I] := 0;
      // try to read data from device
      ReadFile( DeviceInfo.Handle,
                ReadBuffer,
                DeviceInfo.BufferSize,
                BytesRead,
                @HidOverlapped);
      bResult := WaitForSingleObject( ReportEvent, 0);
      If (bResult = WAIT_TIMEOUT) OR (bResult = WAIT_ABANDONED) then
        begin
          CancelIo( DeviceInfo.Handle);
          Result := FALSE;
        end
      else
        begin
          // read count of bytes actually read
          GetOverlappedResult( DeviceInfo.Handle,
                               HidOverlapped,
                               BytesRead,
                               FALSE);
          // copy data cotent to user data structure
          Data := ReadBuffer.Data;
          Result := TRUE;
        end;
    end;
End;

Constructor THIDthread.Create;
begin
  inherited Create( FALSE);
  Priority        := tpTimeCritical;
  FreeOnTerminate := TRUE;
end;

Procedure THIDthread.Free;
Begin
  Inherited;
End;

Procedure THIDthread.Execute;
Var
  Buffer : THIDbuffer;
  I : Integer;
Begin
  While NOT Terminated do  // thread is active (default)
    If (ActiveDevice.Handle <> INVALID_HANDLE_VALUE) then  // a device has been opened
      If Assigned( HIDcallback) then  // callback procedure has been assigned
        begin
          For I := 0 to SizeOf(THIDbuffer) - 1 do Buffer[I] := 0;
          If HIDreadDevice( ActiveDevice,
                            Buffer) then HIDcallBack( Buffer);
          Sleep(1);  // reduce cpu time usage
        end
      else Sleep(100)  // to avoind cpu time usage
    else Sleep(100);  // to avoind cpu time usage
End;

Procedure HIDsetEventHandler( CallBack : THIDcallback);
Begin
  HIDcallback := CallBack;
End;

Constructor TUSBhandler.Create;
Begin
  Inherited Create;
  fMsgHandlerHWND := AllocateHWnd( USBchanger);
End;

Destructor TUSBhandler.Destroy;
Begin
   DeallocateHWnd( fMsgHandlerHWND);
  Inherited Destroy;
End;

procedure TUSBhandler.USBchanger( Var Msg: TMessage);
// this function is triggered when a usb device is attached or removed
begin
  If (Msg.Msg = WM_DEVICECHANGE) then  // usb device inserted or removed
    If Assigned( USBeventHandler) then USBeventHandler;  // call user procedure when defined
  Msg.Result := DefWindowProc( fMsgHandlerHWND,
                               Msg.Msg,
                               Msg.wParam,
                               Msg.lParam);
end;

Procedure USBsetEventHandler( Callback : TUSBcallBack);  stdcall;
Begin
  USBeventHandler := Callback;
End;

INITIALIZATION
  HIDthread           := THIDthread.Create;
  USBhandler          := TUSBhandler.Create( Nil);
  ActiveDevice.Handle := INVALID_HANDLE_VALUE;

FINALIZATION
  USBhandler.Destroy;

END.

